from .generated import ConfigSet as IConfigSet
from ..utils import error_message, is_valid_name


class ConfigSet(IConfigSet):
    """
    A mechanism for organizing and grouping configurations in a hierarchical
    structure. It serves as a path within the configuration tree, enabling
    you to associate specific configuration values with different contexts,
    such as environments, tenants, or any other use case you require.
    """

    SEPARATOR = "/"
    ROOT = ""
    ROOT_LABEL = "/"

    def __init__(self, path: str = None) -> None:
        """
        Creates a new ConfigSet.
        :param: path of the ConfigSet in a hierarchical structure.
        """
        error_location = [self.__class__.__name__, self.__init__.__name__]
        if path is None:
            path = ConfigSet.ROOT
        hierarchy = [ConfigSet.ROOT]
        if path.startswith(ConfigSet.ROOT_LABEL):
            path = path[1:]

        if path.endswith(ConfigSet.SEPARATOR):
            raise ValueError(
                error_message(
                    f"invalid path {path}",
                    error_location,
                    f"path mustn't end with {ConfigSet.SEPARATOR} character",
                )
            )

        for i, step in enumerate(path.split(ConfigSet.SEPARATOR)):
            if not is_valid_name(step):
                raise ValueError(
                    error_message(
                        f"invalid path {path}",
                        error_location,
                        "path is not valid or using reserved name",
                    )
                )
            if step != ConfigSet.ROOT:
                steps = [hierarchy[-1], step] if i > 0 else [step]
                hierarchy.append(ConfigSet.SEPARATOR.join(steps))
        super().__init__(hierarchy=hierarchy, path=path)



from typing import Dict, TypedDict

from ..core import (
    CfguType,
    Command,
    Config,
    ConfigSchema,
    ConfigSet,
    ConfigStore,
)
from ..utils import error_message


class UpsertCommandParameters(TypedDict):
    store: ConfigStore
    set: ConfigSet
    schema: ConfigSchema
    configs: Dict[str, str]


class UpsertCommand(Command):
    """
    The Upsert command is used to create, update or delete Configs from a
    ConfigStore
    """

    parameters: UpsertCommandParameters

    def __init__(
        self,
        *,
        store: ConfigStore,
        set: ConfigSet,
        schema: ConfigSchema,
        configs: Dict[str, str],
    ) -> None:
        """
        Creates a new UpsertCommand.
        :param store: the `configu.core.ConfigStore` to which the command will write
        :param set: the `configu.core.ConfigSet` to which the command will write
        :param schema: `configu.core.ConfigSchema` to validate config being written
        :param configs: a dictionary of configs to upsert
        """
        super().__init__(
            UpsertCommandParameters(
                store=store, set=set, schema=schema, configs=configs
            )
        )

    def run(self):
        """Validates the configs against the schema and upsert to the store

        :raises ValueError: if any config is invalid for the schema
        """
        scope_location = ["UpsertCommand", "run"]
        store = self.parameters["store"]
        set_ = self.parameters["set"]
        schema = self.parameters["schema"]
        configs = self.parameters["configs"]
        store.init()
        schema_content = ConfigSchema.parse(schema)
        upset_configs = []
        for key, value in configs.items():
            cfgu = schema_content.get(key)
            if cfgu is None:
                raise ValueError(
                    error_message(
                        f"invalid config key '{key}'",
                        scope_location,
                        f"key '{key}' must be declared on schema {schema.path}",
                    )
                )
            if value and cfgu.template is not None:
                raise ValueError(
                    error_message(
                        f"invalid assignment to config key '{key}'",
                        scope_location,
                        "keys declared with template mustn't have a value",
                    )
                )
            type_test = ConfigSchema.CFGU.VALIDATORS.get(cfgu.type.value, lambda: False)
            test_values = (
                (
                    value,
                    cfgu.pattern,
                )
                if cfgu.type == CfguType.REG_EX
                else (value,)
            )
            if value and not type_test(*test_values):
                of_type = cfgu.type.value
                if cfgu.type == CfguType.REG_EX:
                    of_type += f"({cfgu.pattern})"
                raise ValueError(
                    error_message(
                        f"invalid config value '{value}' for key '{key}'",
                        scope_location,
                        f"value '{value}' must be of type '{of_type}'",
                    )
                )
            upset_configs.append(Config(set=set_.path, key=key, value=value))
        store.set(upset_configs)



from .core import (
    Config,
    ConfigSchema,
    ConfigSet,
    ConfigStoreQuery,
)
from .stores import (
    AWSSecretsManagerConfigStore,
    AzureKeyVaultConfigStore,
    ConfiguConfigStore,
    ConfiguStoreCredentials,
    GCPSecretManagerConfigStore,
    HashicorpVaultConfigStore,
    InMemoryConfigStore,
    JsonFileConfigStore,
    KubernetesSecretConfigStore,
)
from .commands import (
    EvalCommand,
    ExportCommand,
    UpsertCommand,
)

__doc__ = """
.. include:: ../README.md
.. include:: ../DOCS.md
"""
__all__ = ["core", "stores", "commands"]



from azure.keyvault.secrets import SecretClient
from azure.identity import DefaultAzureCredential
from azure.core.credentials import TokenCredential
from .key_value_store import KeyValueConfigStore


class AzureKeyVaultConfigStore(KeyValueConfigStore):
    """A `ConfigStore` persisted in Azure Key Vault"""

    _client: SecretClient

    def __init__(
        self,
        vault_url: str,
        credentials: TokenCredential = DefaultAzureCredential(),
        **client_kwargs,
    ) -> None:
        """
        :param vault_url: URL of the vault
        :param credentials: A class capable of providing oauth credentials
          (defaults to DefaultAzureCredentials)
        :param kwargs: additional options passed directly to the Azure
          secrets client (see
          https://learn.microsoft.com/en-us/python/api/azure-keyvault-secrets/azure.keyvault.secrets.secretclient?view=azure-python)
        """
        super().__init__(type="azure-key-vault")
        self._client = SecretClient(vault_url, credentials, **client_kwargs)

    def get_by_key(self, key: str) -> str:
        return self._client.get_secret(key).value or ""

    def upsert(self, key: str, value: str):
        self._client.set_secret(key, value)

    def delete(self, key: str):
        self._client.begin_delete_secret(key)



import json
from typing import Dict, List, Union
from abc import abstractmethod

from ..core import Config, ConfigStore, ConfigStoreQuery


class KeyValueConfigStore(ConfigStore):
    """
    An abstract ConfigStore represented by a key-value storage system
    """

    def __init__(self, type: str):
        super().__init__(type)

    @abstractmethod
    def get_by_key(self, key: str) -> str:
        """
        Get data from the store
        :param key: the key in the storage system from which to read
        :return: the value stored in the storage system.
        """
        ...

    @abstractmethod
    def upsert(self, key: str, value: str):
        """
        Upsert data to the storage system
        :param key: the key in the storage system to write to
        :param value: the value to write
        """
        ...

    @abstractmethod
    def delete(self, key: str):
        """
        Delete data from the storage system
        :param key: the key in the storage system to delete
        """
        ...

    def _safe_json_parse(self, value):
        """
        Try parsing the data as json.
        :param value: The value to parse
        :return: The parsed value. If parsing fails, returns an empty dict.
        """
        try:
            return json.loads(value)
        except (Exception,):
            return {}

    def _calc_key(self, query: Union[ConfigStoreQuery, Config]) -> str:
        """
        Calculate the storage system key referenced in a query.
        If `set` is populated, will use that. Otherwise defaults to `key`
        :param query: The query to be calculated.
        """
        return query.set or query.key

    def get(self, queries: List[ConfigStoreQuery]) -> List[Config]:
        stored_configs = []
        for query in queries:
            try:
                value = self.get_by_key(self._calc_key(query))
                if query.set:
                    value = str(self._safe_json_parse(value).get(query.key, ""))
                if value:
                    stored_configs.append(
                        Config.from_dict(
                            {
                                "key": query.key,
                                "set": query.set,
                                "value": value,
                            }
                        )
                    )
            except (Exception,):
                continue
        return stored_configs

    def set(self, configs: List[Config]):
        key_value_dict: Dict[str, Union[str, Dict[str, str]]] = {}
        for config in configs:
            key = self._calc_key(config)
            if not config.set:
                key_value_dict[key] = config.value
                continue

            if key not in key_value_dict or not isinstance(key_value_dict[key], dict):
                key_value_dict[key] = {}

            key_value_dict[key] = {
                **key_value_dict[key],
                config.key: config.value,
            }

        for key, value in key_value_dict.items():
            if value == "" or value is None:
                self.delete(key)
            else:
                self.upsert(
                    key,
                    json.dumps(value) if isinstance(value, dict) else str(value),
                )



import os
import json
import requests
from .key_value_store import KeyValueConfigStore


class HashicorpVaultConfigStore(KeyValueConfigStore):
    """A `ConfigStore` persisted in Hashicorp Vault"""

    def __init__(
        self,
        engine,
        address=os.environ.get("VAULT_ADDR"),
        token=os.environ.get("VAULT_TOKEN"),
    ):
        """
        :param engine: The secrets engine to use
        :param address: The vault address (defaults to VAULT_ADDR environment
            variable)
        :param token: The vault token (defaults to VAULT_TOKEN environment
            variable)
        """
        self.engine = engine
        self.address = address
        self.token = token
        super().__init__(type="hashicorp-vault")

    def _format_key(self, key: str) -> str:
        return f"{self.engine}/data/{key}"

    def get_by_key(self, key: str) -> str:
        return requests.get(
            f"{self.address}/v1/{self._format_key(key)}",
            headers={
                "X-Vault-Token": self.token,
            },
        ).json()

    def delete(self, key: str):
        metadata_key = self._format_key(key).replace("data", "metadata", 1)
        requests.delete(
            f"{self.address}/v1/{metadata_key}",
            headers={
                "X-Vault-Token": self.token,
            },
        )

    def upsert(self, key: str, value: str):
        requests.post(
            f"{self.address}/v1/{self._format_key(key)}",
            headers={
                "X-Vault-Token": self.token,
            },
            json={"data": json.loads(value)},
        )



from dataclasses import dataclass
from typing import Dict, List

import pyvalidator
import requests
from requests.auth import AuthBase

from ..core import Config, ConfigStore, ConfigStoreQuery


@dataclass
class ConfiguStoreCredentials:
    org: str
    token: str


class BearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class TokenAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["token"] = self.token
        return r


class ConfiguConfigStore(ConfigStore):
    """A `ConfigStore` persisted by Configu (https://app.configu.com)"""

    _headers: Dict[str, str]
    _auth: AuthBase
    _url: str

    def __init__(
        self,
        credentials: ConfiguStoreCredentials,
        *,
        endpoint: str = "https://api.configu.com",
        source: str = "sdk",
    ) -> None:
        super().__init__(type="configu")
        self._headers = {
            "Org": credentials.org,
            "Source": source,
        }
        token = credentials.token
        self._auth = (
            BearerAuth(token) if pyvalidator.is_jwt(token) else TokenAuth(token)
        )
        self._url = f"{endpoint}/config"

    def get(self, queries: List[ConfigStoreQuery]) -> List[Config]:
        queries_json = {"queries": [query.to_dict() for query in queries]}
        response = requests.post(
            url=self._url,
            headers=self._headers,
            auth=self._auth,
            json=queries_json,
        )
        response.raise_for_status()
        return [Config(**args) for args in response.json()]

    def set(self, configs: List[Config]) -> None:
        configs_json = {"configs": [config.to_dict() for config in configs]}
        response = requests.put(
            url=self._url,
            headers=self._headers,
            auth=self._auth,
            json=configs_json,
        )
        response.raise_for_status()



from typing import List

from ..core import Config, ConfigStore, ConfigStoreQuery


class InMemoryConfigStore(ConfigStore):
    """A `ConfigStore` persisted in-memory"""

    _data: List[Config]

    def __init__(self) -> None:
        super().__init__(type="in-memory")
        self._data = []

    def get(self, queries: List[ConfigStoreQuery]) -> List[Config]:
        query_ids = [f"{query.set}.{query.key}" for query in queries]
        return [
            config for config in self._data if f"{config.set}.{config.key}" in query_ids
        ]

    def set(self, configs: List[Config]) -> None:
        set_config_ids = [f"{config.set}.{config.key}" for config in configs]
        existing = [
            config
            for config in self._data
            if f"{config.set}.{config.key}" not in set_config_ids
        ]
        self._data = [config for config in existing + configs if bool(config.value)]



import boto3
from mypy_boto3_secretsmanager import SecretsManagerClient
from .key_value_store import KeyValueConfigStore


class AWSSecretsManagerConfigStore(KeyValueConfigStore):
    """A `ConfigStore` persisted in AWS Secrets Manager"""

    _client: SecretsManagerClient

    def __init__(self, **client_kwargs) -> None:
        super().__init__(type="aws-secrets-manager")
        self._client = boto3.client("secretsmanager", **client_kwargs)

    def get_by_key(self, key: str) -> str:
        return self._client.get_secret_value(SecretId=key).get("SecretString", "")

    def upsert(self, key: str, value: str):
        try:
            self._client.update_secret(SecretId=key, SecretString=value)
        except (Exception,):
            self._client.create_secret(Name=key, SecretString=value)

    def delete(self, key: str):
        self._client.delete_secret(SecretId=key)



import json
from typing import List

from ..core import Config, ConfigStore, ConfigStoreQuery


class JsonFileConfigStore(ConfigStore):
    """A `ConfigStore` persisted in a json file"""

    _path: str

    def __init__(self, path: str) -> None:
        super().__init__(type="json-file")
        self._path = path

    def _read(self) -> List[Config]:
        """
        Reads JSON file contents
        :return: The file contents, represented as a list of Config
        dicts
        """
        with open(self._path, mode="r", encoding="utf-8") as json_file:
            data = json.load(json_file)
        return [Config(**args) for args in data]

    def _write(self, configs: List[Config]) -> None:
        """
        Writes to the JSON file
        :param configs: a list of key/set/value dicts to write
        """
        data = [config.to_dict() for config in configs]
        with open(self._path, mode="w", encoding="utf-8") as json_file:
            json.dump(data, json_file)

    def get(self, queries: List[ConfigStoreQuery]) -> List[Config]:
        stored_configs = self._read()
        query_ids = [f"{query.set}.{query.key}" for query in queries]
        return [
            config
            for config in stored_configs
            if f"{config.set}.{config.key}" in query_ids
        ]

    def set(self, configs: List[Config]) -> None:
        stored_configs = self._read()
        set_config_ids = [f"{config.set}.{config.key}" for config in configs]
        existing = [
            config
            for config in stored_configs
            if f"{config.set}.{config.key}" not in set_config_ids
        ]
        next_configs = [config for config in existing + configs if bool(config.value)]
        self._write(next_configs)



from abc import ABC, abstractmethod
from typing import List

from .generated import Config, ConfigStore as IConfigStore, ConfigStoreQuery


class ConfigStore(ABC, IConfigStore):
    """
    An abstract base class representing a config store.

    A ConfigStore is a storage system for Configs in Configu,
    serving as a mechanism for storing, retrieving, and managing configuration
    data.
    """

    def init(self):
        """Creates/prepares a new ConfigStore"""

    @abstractmethod
    def get(self, queries: List[ConfigStoreQuery]) -> List[Config]:
        """
        Retrieves a set of `Config`s.
        :param queries: a class containing either the key and/or set to be
        retrieved.
        :return: A list of `Config`s from the store.
        """

    @abstractmethod
    def set(self, configs: List[Config]) -> None:
        """
        Upserts a set of `Config`s.
        :param configs: a list of `Config`s to be set in the store.
        """



import os
from typing import Dict, TypedDict

from .eval_command import EvalCommandReturn
from ..core import (
    Command,
)


class ExportCommandParameters(TypedDict):
    data: EvalCommandReturn
    env: bool
    override: bool


ExportCommandReturn = Dict[str, str]


class ExportCommand(Command[ExportCommandReturn]):
    """"""

    parameters: ExportCommandParameters

    def __init__(
        self,
        data: EvalCommandReturn,
        *,
        env: bool = True,
        override: bool = True,
    ) -> None:
        super().__init__(ExportCommandParameters(data=data, env=env, override=override))

    def run(self) -> ExportCommandReturn:
        data = self.parameters["data"]
        env = self.parameters.get("env", True)
        override = self.parameters.get("override", True)
        exported_configs: ExportCommandReturn = {
            key: value["result"]["value"] for key, value in data.items()
        }
        if env:
            for key, value in exported_configs.items():
                if override or key in os.environ:
                    os.environ[key] = value

        return exported_configs



import json
import re
from itertools import cycle
from pathlib import Path
from typing import Callable, Dict

import pyvalidator

from .generated import (
    Cfgu,
    CfguType,
    ConfigSchema as IConfigSchema,
    ConfigSchemaType,
)
from ..utils import error_message, is_valid_name


class ConfigSchemaDefinition:
    _cs_regex = r"^(?:([^:/?#\s]+):/{2})?(?:([^@/?#\s]+)@)?([^/?#\s]+)?(?:/([^?#\s]*))?(?:[?]([^#\s]+))?\S*$"  # noqa: E501
    NAME: str = "cfgu"
    EXT: str = ".cfgu"
    VALIDATORS: Dict[str, Callable[[str], bool]] = {
        "Boolean": pyvalidator.is_boolean,
        "Number": pyvalidator.is_number,
        "String": lambda value: isinstance(value, str),
        "RegEx": lambda *args: re.fullmatch(args[1], args[0]) is not None,
        "UUID": pyvalidator.is_uuid,
        "SemVer": pyvalidator.is_semantic_version,
        "Email": pyvalidator.is_email,
        "MobilePhone": pyvalidator.is_mobile_number,
        "LatLong": pyvalidator.is_lat_long,
        "Color": lambda value: (
            pyvalidator.is_hexadecimal(value)
            or pyvalidator.is_hsl(value)
            or pyvalidator.is_rgb_color(value)
        ),
        "IPv4": lambda value: pyvalidator.is_ip(value, 4),
        "IPv6": lambda value: pyvalidator.is_ip(value, 6),
        "Domain": pyvalidator.is_fqdn,
        "URL": pyvalidator.is_url,
        "ConnectionString": lambda value: re.fullmatch(
            ConfigSchemaDefinition._cs_regex,
            value,
            re.RegexFlag.M,
        )
        is not None,
        "Hex": pyvalidator.is_hexadecimal,
        "Base64": pyvalidator.is_base64,
        "MD5": pyvalidator.is_md5,
        "SHA": lambda value: (
            pyvalidator.is_hash(value, "sha1")
            or pyvalidator.is_hash(value, "sha256")
            or pyvalidator.is_hash(value, "sha384")
            or pyvalidator.is_hash(value, "sha512")
        ),
        "Currency": pyvalidator.is_currency,
    }
    PROPS = list(Cfgu.__annotations__.keys())


class ConfigSchema(IConfigSchema):
    """
    A Python representation of the schema stored in a .cfgu.json file.
    """

    CFGU = ConfigSchemaDefinition

    TYPES = {f".{schema_type.value}": schema_type for schema_type in ConfigSchemaType}
    EXT = " | ".join(["".join(ext) for ext in zip(cycle(CFGU.EXT), TYPES.keys())])

    def __init__(self, path: str) -> None:
        """
        Creates a new ConfigSchema
        :param path: path to the schema file (.cfgu.json)
        """
        error_location = [self.__class__.__name__, self.__init__.__name__]
        if re.match(rf".*({ConfigSchema.EXT})", path) is None:
            raise ValueError(
                error_message(
                    f"invalid path {path}",
                    error_location,
                    f"file extension must be {ConfigSchema.EXT}",
                )
            )
        super().__init__(path=path, type=ConfigSchema.TYPES[Path(path).suffix])

    def read(self) -> str:
        """
        Reads the config schema file
        :return: contents of the config schema
        """
        with open(self.path, mode="r", encoding="utf-8") as schema_file:
            file_content = schema_file.read()
        return file_content

    @classmethod
    def parse(cls, scheme: "ConfigSchema") -> Dict[str, Cfgu]:
        """
        Parses the given ConfigSchema
        :param scheme: The ConfigSchema to parse
        :return: A dictionary of configurations.
        """
        error_location = [cls.__name__, "parse"]
        schema_content = scheme.read()
        if scheme.type == ConfigSchemaType.JSON:
            schema_content = json.loads(schema_content)
            assert isinstance(schema_content, dict)
            schema_content = {k: Cfgu.from_dict(v) for (k, v) in schema_content.items()}

        # validate parsed
        for key, cfgu in schema_content.items():
            if not is_valid_name(key):
                raise ValueError(
                    error_message(
                        f"invalid key {key}",
                        error_location + [key],
                        f"path nodes mustn't contain reserved words '${key}'",
                    )
                )
            if cfgu.type == CfguType.REG_EX and cfgu.pattern is None:
                raise ValueError(
                    error_message(
                        "invalid type property",
                        error_location + [key, cfgu.type.value],
                    ),
                    f"type '{cfgu.type.value}' must come with a pattern property",
                )
            if cfgu.default is not None:
                if cfgu.required is not None or cfgu.template is not None:
                    raise ValueError(
                        error_message(
                            "invalid default property",
                            error_location + [key, "default"],
                        ),
                        "default mustn't set together with required "
                        "or template properties",
                    )
                else:
                    type_test = ConfigSchema.CFGU.VALIDATORS.get(
                        cfgu.type.value, lambda: False
                    )
                    test_values = (
                        (cfgu.default, cfgu.pattern)
                        if cfgu.type == CfguType.REG_EX
                        else (cfgu.default,)
                    )
                    if not type_test(*test_values):
                        raise ValueError(
                            error_message(
                                "invalid default property",
                                error_location + [key, "default"],
                            ),
                            f"{cfgu.default} must be of type {cfgu.type.value}"
                            f" or match Regex",
                        )

            if cfgu.depends is not None and (
                not len(cfgu.depends)
                or any([not is_valid_name(dependency) for dependency in cfgu.depends])
            ):
                raise ValueError(
                    error_message(
                        "invalid depends property",
                        error_location + [key, "depends"],
                    ),
                    "depends is empty or contain reserved words",
                )

        return schema_content



from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

CommandReturn = TypeVar("CommandReturn")


@dataclass
class Command(ABC, Generic[CommandReturn]):
    """
    Abstract base class encapsulating a Configu command.
    """

    parameters: Mapping[str, Any]

    @abstractmethod
    def run(self) -> CommandReturn:
        """
        Overridden by individual commands.
        Contains instructions for how to run a command.
        """



from kubernetes import client, config
import base64
import json
from .key_value_store import KeyValueConfigStore


class KubernetesSecretConfigStore(KeyValueConfigStore):
    """A `ConfigStore` persisted in Kubernetes Secrets"""

    _client: client.CoreV1Api
    _namespace: str

    def __init__(self, namespace: str, kubeconfig=None) -> None:
        """
        :param namespace: The namespace of your secrets
        :param kubeconfig: (Optional) Path to your kube-config file.
        """
        config.load_kube_config(kubeconfig)
        self._client = client.CoreV1Api()
        self._namespace = namespace
        super().__init__(type="kubernetes-secret")

    def get_by_key(self, key: str) -> str:
        response = self._client.read_namespaced_secret(key, self._namespace)
        response_dict = response.data
        response_dict = {
            k: base64.b64decode(v.encode("utf-8")).decode("utf-8")
            for k, v in response_dict.items()
        }
        return json.dumps(response_dict)

    def upsert(self, key: str, value: str):
        value_dict = json.loads(value)
        value_dict = {
            k: base64.b64encode(v.encode("utf-8")).decode("utf-8")
            for k, v in value_dict.items()
        }
        try:
            self._client.create_namespaced_secret(
                self._namespace,
                {
                    "metadata": {"name": key},
                    "data": value_dict,
                },
            )
        except (Exception,):
            self._client.patch_namespaced_secret(
                key,
                self._namespace,
                {
                    "data": value_dict,
                },
            )

    def delete(self, key: str):
        self._client.delete_namespaced_secret(key, self._namespace)



from .aws_secrets_manager_store import AWSSecretsManagerConfigStore
from .configu_store import ConfiguConfigStore, ConfiguStoreCredentials
from .in_memory_store import InMemoryConfigStore
from .json_file_store import JsonFileConfigStore
from .hashicorp_vault_store import HashicorpVaultConfigStore
from .azure_key_vault_store import AzureKeyVaultConfigStore
from .gcp_secret_manager import GCPSecretManagerConfigStore
from .kubernetes_secret_store import KubernetesSecretConfigStore

__all__ = [
    "AWSSecretsManagerConfigStore",
    "InMemoryConfigStore",
    "ConfiguConfigStore",
    "ConfiguStoreCredentials",
    "JsonFileConfigStore",
    "HashicorpVaultConfigStore",
    "AzureKeyVaultConfigStore",
    "GCPSecretManagerConfigStore",
    "KubernetesSecretConfigStore",
]



import re
from typing import Any, Dict, List, Optional

import chevron


def error_message(
    message: str,
    location: Optional[List[str]] = None,
    suggestion: Optional[str] = None,
) -> str:
    location = f"at {' > '.join(location)}" if location else None
    return " ".join(
        [detail for detail in [message, location, suggestion] if detail is not None]
    )


def is_valid_name(name: str) -> bool:
    naming_pattern = r"^[A-Za-z0-9_-]*$"
    reserved_names = ["_", "-", "this", "cfgu"]
    return name not in reserved_names and re.match(naming_pattern, name) is not None


def parse_template(template: str) -> List[str]:
    template_vars = []
    for tag_type, tag_key in chevron.tokenizer.tokenize(template):
        if tag_type not in ["variable", "literal"]:
            raise Exception(
                error_message(f'invalid template "{template}"', ["Template", "parse"])
            )
        if tag_type == "variable":
            template_vars.append(tag_key)
    return template_vars


def render_template(template: str, context: Dict[str, Any]) -> str:
    return chevron.render(template, context)



from .eval_command import (
    EvalCommand,
    EvalCommandParameters,
    EvalCommandReturn,
    EvalCommandReturnContext,
    EvalCommandReturnResult,
    EvalCommandReturnValue,
)
from .upsert_command import UpsertCommand, UpsertCommandParameters
from .export_command import (
    ExportCommand,
    ExportCommandParameters,
    ExportCommandReturn,
)

__all__ = [
    "EvalCommand",
    "EvalCommandParameters",
    "EvalCommandReturn",
    "EvalCommandReturnContext",
    "EvalCommandReturnResult",
    "EvalCommandReturnValue",
    "UpsertCommand",
    "UpsertCommandParameters",
    "ExportCommand",
    "ExportCommandParameters",
    "ExportCommandReturn",
]



from enum import Enum
from functools import reduce
from typing import Dict, Tuple, TypedDict, Optional

from ..core import (
    Cfgu,
    CfguType,
    Command,
    Config,
    ConfigSchema,
    ConfigSet,
    ConfigStore,
    ConfigStoreQuery,
)
from ..utils import error_message, parse_template, render_template


class EvaluatedConfigOrigin(Enum):
    ConfigsOverride = "CONFIGS_OVERRIDE"
    StoreSet = "STORE_SET"
    SchemaTemplate = "SCHEMA_TEMPLATE"
    SchemaDefault = "SCHEMA_DEFAULT"
    EmptyValue = "EMPTY_VALUE"


class EvalCommandReturnContext(TypedDict):
    store: str
    set: str
    schema: str
    key: str
    cfgu: Cfgu


class EvalCommandReturnResult(TypedDict):
    origin: EvaluatedConfigOrigin
    source: str
    value: str


class EvalCommandReturnValue(TypedDict):
    context: EvalCommandReturnContext
    result: EvalCommandReturnResult


EvalCommandReturn = Dict[str, EvalCommandReturnValue]


class EvalCommandParameters(TypedDict):
    store: ConfigStore
    set: ConfigSet
    schema: ConfigSchema
    configs: Optional[Dict[str, str]]
    validate: bool
    previous: Optional[EvalCommandReturn]


class EvalCommand(Command[EvalCommandReturn]):
    """
    The Eval command is used to fetch and validate `Config`s from ConfigStore
    on demand.
    """

    parameters: EvalCommandParameters

    def __init__(
        self,
        *,
        store: ConfigStore,
        set: ConfigSet,
        schema: ConfigSchema,
        configs: Dict[str, str] = None,
        validate: bool = True,
        previous: EvalCommandReturn = None,
    ) -> None:
        """

        :param store: the `configu.core.ConfigStore` from which to fetch
        :param set: the `ConfigSet` to fetch
        :param schema: `ConfigSchema` to validate the config being fetched
        :param configs: a dictionary of overrides to the fetched Config`s
        :param validate: run validation against schema, defaults to True
        :param previous: the previous `EvalCommandReturn` in case of pipes
        """
        super().__init__(
            EvalCommandParameters(
                store=store,
                set=set,
                schema=schema,
                configs=configs,
                validate=validate,
                previous=previous,
            )
        )

    def _eval_from_configs_override(
        self, result: EvalCommandReturn
    ) -> EvalCommandReturn:
        if not self.parameters.get("configs"):
            return {}
        for key, value in result.items():
            if key in self.parameters["configs"]:
                override_value = self.parameters["configs"][key]
                value["result"]["origin"] = EvaluatedConfigOrigin.ConfigsOverride
                value["result"]["source"] = f"parameters.configs.{key}={override_value}"
                value["result"]["value"] = override_value
        return result

    def _eval_from_store_set(self, result: EvalCommandReturn) -> EvalCommandReturn:
        store = self.parameters["store"]
        set_ = self.parameters["set"]
        store_queries = [
            ConfigStoreQuery(key, store_set)
            for store_set in set_.hierarchy
            for key in result
        ]
        store_configs: Dict[str, Config] = {
            result.key: result
            for result in sorted(
                store.get(store_queries),
                key=lambda query_result: len(query_result.set.split(set_.SEPARATOR)),
            )
        }
        for key, value in result.items():
            if key in store_configs:
                store_config = store_configs[key]
                value["result"]["origin"] = EvaluatedConfigOrigin.StoreSet
                value["result"]["source"] = (
                    f"parameters.store={value['context']['store']},"
                    f"parameters.set={value['context']['set']}"
                )
                value["result"]["value"] = store_config.value
        return result

    @staticmethod
    def _eval_from_schema(result: EvalCommandReturn) -> EvalCommandReturn:
        for key, value in result.items():
            context = value["context"]
            cfgu = context["cfgu"]
            if cfgu.template:
                value["result"]["origin"] = EvaluatedConfigOrigin.SchemaTemplate
                value["result"]["source"] = (
                    f"parameters.schema={context['schema']}"
                    f".template={cfgu.template}"
                )
                value["result"]["value"] = ""

            if cfgu.default:
                value["result"]["origin"] = EvaluatedConfigOrigin.SchemaDefault
                value["result"]["source"] = (
                    f"parameters.schema={context['schema']}" f".default={cfgu.default}"
                )
                value["result"]["value"] = cfgu.default
        return result

    def _validate_result(self, result: EvalCommandReturn):
        if self.parameters.get("validate", True):
            error_scope = ["EvalCommand", "run"]
            for key, value in result.items():
                cfgu = value["context"]["cfgu"]
                evaluated_value = value["result"]["value"]
                type_test = ConfigSchema.CFGU.VALIDATORS.get(
                    cfgu.type.value, lambda: False
                )
                test_values = (
                    (
                        evaluated_value,
                        cfgu.pattern,
                    )
                    if cfgu.type == CfguType.REG_EX
                    else (evaluated_value,)
                )
                if not type_test(*test_values):
                    raise ValueError(
                        error_message(
                            f"invalid value type for key '{key}'", error_scope
                        ),
                        f"value '{test_values[0]}' must be a " f"'{cfgu.type}'",
                    )
                if cfgu.required is not None and not bool(test_values[0]):
                    raise ValueError(
                        error_message(
                            f"required key '{key}' is missing a value",
                            error_scope,
                        )
                    )
                if bool(test_values[0]) and cfgu.depends is not None:
                    if any(
                        [
                            True
                            for dep in cfgu.depends
                            if dep not in result.keys()
                            or not bool(result[dep]["result"]["value"])
                        ]
                    ):
                        raise ValueError(
                            error_message(
                                f"one or more depends of key '{key}'"
                                f" is missing a value",
                                error_scope,
                            )
                        )

    def _eval_previous(self, result: EvalCommandReturn) -> EvalCommandReturn:
        previous_result = self.parameters.get("previous")
        if not previous_result:
            return {}

        def reduce_prev(
            merged: EvalCommandReturn,
            current: Tuple[str, EvalCommandReturnValue],
        ):
            key, value = current
            if key not in merged or (
                merged[key]["result"]["origin"] == EvaluatedConfigOrigin.EmptyValue
                and value["result"]["origin"] != EvaluatedConfigOrigin.EmptyValue
            ):
                merged[key] = value
            return merged

        return reduce(
            reduce_prev,
            reversed(list(previous_result.items()) + list(result.items())),
            {},
        )

    @staticmethod
    def _eval_templates(result: EvalCommandReturn) -> EvalCommandReturn:
        template_keys = list(
            {
                key
                for key, value in result.items()
                if value["result"]["origin"] == EvaluatedConfigOrigin.SchemaTemplate
            }
        )
        should_render_templates = True
        while len(template_keys) and should_render_templates:
            has_rendered_at_least_once = False
            for key in template_keys:
                context = result[key]["context"]
                template = context["cfgu"].template
                expressions = parse_template(template)
                if any([True for exp in expressions if exp in template_keys]):
                    continue
                context_config_set = ConfigSet(context["set"])
                render_context = {
                    **{key: value["result"]["value"] for key, value in result.items()},
                    "CONFIGU_STORE": {"type": context["store"]},
                    **{
                        "CONFIGU_SET": {
                            "path": context_config_set.path,
                            "hierarchy": context_config_set.hierarchy,
                            "first": context_config_set.hierarchy[0],
                            "last": context_config_set.hierarchy[-1],
                            **{
                                str(index): path
                                for index, path in enumerate(
                                    context_config_set.hierarchy
                                )
                            },
                        }
                    },
                    "CONFIGU_SCHEMA": {"path": context["schema"]},
                }
                result[key]["result"]["value"] = render_template(
                    template, render_context
                )
                template_keys.remove(key)
                has_rendered_at_least_once = True
            should_render_templates = has_rendered_at_least_once
        return result

    def run(self):
        """
        Runs the eval command.

        :return: `EvalCommandReturn`
        The evaluated configs contains the command's
        results and metadata

        :raises: AnyError If anything bad happens.
        """
        store = self.parameters["store"]
        set_ = self.parameters["set"]
        schema = self.parameters["schema"]
        store.init()
        schema_contents = ConfigSchema.parse(schema)
        result: EvalCommandReturn = {
            key: {
                "context": {
                    "store": store.type,
                    "set": set_.path,
                    "schema": schema.path,
                    "key": key,
                    "cfgu": cfgu,
                },
                "result": {
                    "origin": EvaluatedConfigOrigin.EmptyValue,
                    "source": "",
                    "value": "",
                },
            }
            for key, cfgu in schema_contents.items()
        }

        result = {**result, **self._eval_from_configs_override(result)}

        result = {
            **result,
            **self._eval_from_store_set(
                {
                    key: value
                    for key, value in result.items()
                    if value["result"]["origin"] == EvaluatedConfigOrigin.EmptyValue
                    and not value["context"]["cfgu"].template
                }
            ),
        }

        result = {
            **result,
            **self._eval_from_schema(
                {
                    key: value
                    for key, value in result.items()
                    if value["result"]["origin"] == EvaluatedConfigOrigin.EmptyValue
                }
            ),
        }

        result = {**result, **self._eval_previous(result)}
        result = {**result, **self._eval_templates(result)}
        self._validate_result(result)

        return result



from enum import Enum
from dataclasses import dataclass
from typing import Optional, List, Any, Dict, TypeVar, Callable, Type, cast


T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def to_enum(c: Type[EnumT], x: Any) -> EnumT:
    assert isinstance(x, c)
    return x.value


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


class CfguType(Enum):
    ALIBABA_REGION = "AlibabaRegion"
    AWS_REGION = "AwsRegion"
    AZ_REGION = "AZRegion"
    BASE64 = "Base64"
    BOOLEAN = "Boolean"
    COLOR = "Color"
    CONNECTION_STRING = "ConnectionString"
    COUNTRY = "Country"
    CURRENCY = "Currency"
    DOCKER_IMAGE = "DockerImage"
    DOMAIN = "Domain"
    EMAIL = "Email"
    GCP_REGION = "GCPRegion"
    HEX = "Hex"
    IBM_REGION = "IBMRegion"
    I_PV4 = "IPv4"
    I_PV6 = "IPv6"
    LANGUAGE = "Language"
    LAT_LONG = "LatLong"
    LOCALE = "Locale"
    MAC_ADDRESS = "MACAddress"
    MD5 = "MD5"
    MIME_TYPE = "MIMEType"
    MOBILE_PHONE = "MobilePhone"
    NUMBER = "Number"
    ORACLE_REGION = "OracleRegion"
    REG_EX = "RegEx"
    SEM_VER = "SemVer"
    SHA = "SHA"
    STRING = "String"
    URL = "URL"
    UUID = "UUID"


@dataclass
class Cfgu:
    """A generic declaration of a Config, aka Cfgu that specifies information about its type and
    other characteristics
    """
    type: CfguType
    default: Optional[str] = None
    depends: Optional[List[str]] = None
    description: Optional[str] = None
    pattern: Optional[str] = None
    required: Optional[bool] = None
    template: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Cfgu':
        assert isinstance(obj, dict)
        type = CfguType(obj.get("type"))
        default = from_union([from_str, from_none], obj.get("default"))
        depends = from_union([lambda x: from_list(from_str, x), from_none], obj.get("depends"))
        description = from_union([from_str, from_none], obj.get("description"))
        pattern = from_union([from_str, from_none], obj.get("pattern"))
        required = from_union([from_bool, from_none], obj.get("required"))
        template = from_union([from_str, from_none], obj.get("template"))
        return Cfgu(type, default, depends, description, pattern, required, template)

    def to_dict(self) -> dict:
        result: dict = {}
        result["type"] = to_enum(CfguType, self.type)
        if self.default is not None:
            result["default"] = from_union([from_str, from_none], self.default)
        if self.depends is not None:
            result["depends"] = from_union([lambda x: from_list(from_str, x), from_none], self.depends)
        if self.description is not None:
            result["description"] = from_union([from_str, from_none], self.description)
        if self.pattern is not None:
            result["pattern"] = from_union([from_str, from_none], self.pattern)
        if self.required is not None:
            result["required"] = from_union([from_bool, from_none], self.required)
        if self.template is not None:
            result["template"] = from_union([from_str, from_none], self.template)
        return result


@dataclass
class Config:
    """A generic representation of a software configuration, aka Config"""
    key: str
    set: str
    value: str

    @staticmethod
    def from_dict(obj: Any) -> 'Config':
        assert isinstance(obj, dict)
        key = from_str(obj.get("key"))
        set = from_str(obj.get("set"))
        value = from_str(obj.get("value"))
        return Config(key, set, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["key"] = from_str(self.key)
        result["set"] = from_str(self.set)
        result["value"] = from_str(self.value)
        return result


class ConfigSchemaType(Enum):
    JSON = "json"


@dataclass
class ConfigSchema:
    """An interface of a <file>.cfgu.json, aka ConfigSchema
    that contains binding records between a unique Config.<key> and its Cfgu declaration
    """
    path: str
    type: ConfigSchemaType

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigSchema':
        assert isinstance(obj, dict)
        path = from_str(obj.get("path"))
        type = ConfigSchemaType(obj.get("type"))
        return ConfigSchema(path, type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["path"] = from_str(self.path)
        result["type"] = to_enum(ConfigSchemaType, self.type)
        return result


@dataclass
class ConfigSchemaContentsValue:
    type: CfguType
    default: Optional[str] = None
    depends: Optional[List[str]] = None
    description: Optional[str] = None
    pattern: Optional[str] = None
    required: Optional[bool] = None
    template: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigSchemaContentsValue':
        assert isinstance(obj, dict)
        type = CfguType(obj.get("type"))
        default = from_union([from_str, from_none], obj.get("default"))
        depends = from_union([lambda x: from_list(from_str, x), from_none], obj.get("depends"))
        description = from_union([from_str, from_none], obj.get("description"))
        pattern = from_union([from_str, from_none], obj.get("pattern"))
        required = from_union([from_bool, from_none], obj.get("required"))
        template = from_union([from_str, from_none], obj.get("template"))
        return ConfigSchemaContentsValue(type, default, depends, description, pattern, required, template)

    def to_dict(self) -> dict:
        result: dict = {}
        result["type"] = to_enum(CfguType, self.type)
        if self.default is not None:
            result["default"] = from_union([from_str, from_none], self.default)
        if self.depends is not None:
            result["depends"] = from_union([lambda x: from_list(from_str, x), from_none], self.depends)
        if self.description is not None:
            result["description"] = from_union([from_str, from_none], self.description)
        if self.pattern is not None:
            result["pattern"] = from_union([from_str, from_none], self.pattern)
        if self.required is not None:
            result["required"] = from_union([from_bool, from_none], self.required)
        if self.template is not None:
            result["template"] = from_union([from_str, from_none], self.template)
        return result


@dataclass
class ConfigSet:
    """An interface of a path in an hierarchy, aka ConfigSet
    that uniquely groups Config.<key> with their Config.<value>.
    """
    hierarchy: List[str]
    path: str

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigSet':
        assert isinstance(obj, dict)
        hierarchy = from_list(from_str, obj.get("hierarchy"))
        path = from_str(obj.get("path"))
        return ConfigSet(hierarchy, path)

    def to_dict(self) -> dict:
        result: dict = {}
        result["hierarchy"] = from_list(from_str, self.hierarchy)
        result["path"] = from_str(self.path)
        return result


@dataclass
class ConfigStore:
    """An interface of a storage, aka ConfigStore
    that I/Os Config records (Config[])
    """
    type: str

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigStore':
        assert isinstance(obj, dict)
        type = from_str(obj.get("type"))
        return ConfigStore(type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["type"] = from_str(self.type)
        return result


@dataclass
class ConfigStoreQuery:
    key: str
    set: str

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigStoreQuery':
        assert isinstance(obj, dict)
        key = from_str(obj.get("key"))
        set = from_str(obj.get("set"))
        return ConfigStoreQuery(key, set)

    def to_dict(self) -> dict:
        result: dict = {}
        result["key"] = from_str(self.key)
        result["set"] = from_str(self.set)
        return result


@dataclass
class ConfigStoreContentsElement:
    key: str
    set: str
    value: str

    @staticmethod
    def from_dict(obj: Any) -> 'ConfigStoreContentsElement':
        assert isinstance(obj, dict)
        key = from_str(obj.get("key"))
        set = from_str(obj.get("set"))
        value = from_str(obj.get("value"))
        return ConfigStoreContentsElement(key, set, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["key"] = from_str(self.key)
        result["set"] = from_str(self.set)
        result["value"] = from_str(self.value)
        return result


def cfgu_type_from_dict(s: Any) -> CfguType:
    return CfguType(s)


def cfgu_type_to_dict(x: CfguType) -> Any:
    return to_enum(CfguType, x)


def cfgu_from_dict(s: Any) -> Cfgu:
    return Cfgu.from_dict(s)


def cfgu_to_dict(x: Cfgu) -> Any:
    return to_class(Cfgu, x)


def config_from_dict(s: Any) -> Config:
    return Config.from_dict(s)


def config_to_dict(x: Config) -> Any:
    return to_class(Config, x)


def config_schema_type_from_dict(s: Any) -> ConfigSchemaType:
    return ConfigSchemaType(s)


def config_schema_type_to_dict(x: ConfigSchemaType) -> Any:
    return to_enum(ConfigSchemaType, x)


def config_schema_from_dict(s: Any) -> ConfigSchema:
    return ConfigSchema.from_dict(s)


def config_schema_to_dict(x: ConfigSchema) -> Any:
    return to_class(ConfigSchema, x)


def config_schema_contents_value_from_dict(s: Any) -> ConfigSchemaContentsValue:
    return ConfigSchemaContentsValue.from_dict(s)


def config_schema_contents_value_to_dict(x: ConfigSchemaContentsValue) -> Any:
    return to_class(ConfigSchemaContentsValue, x)


def config_schema_contents_from_dict(s: Any) -> Dict[str, ConfigSchemaContentsValue]:
    return from_dict(ConfigSchemaContentsValue.from_dict, s)


def config_schema_contents_to_dict(x: Dict[str, ConfigSchemaContentsValue]) -> Any:
    return from_dict(lambda x: to_class(ConfigSchemaContentsValue, x), x)


def config_set_from_dict(s: Any) -> ConfigSet:
    return ConfigSet.from_dict(s)


def config_set_to_dict(x: ConfigSet) -> Any:
    return to_class(ConfigSet, x)


def config_store_from_dict(s: Any) -> ConfigStore:
    return ConfigStore.from_dict(s)


def config_store_to_dict(x: ConfigStore) -> Any:
    return to_class(ConfigStore, x)


def config_store_query_from_dict(s: Any) -> ConfigStoreQuery:
    return ConfigStoreQuery.from_dict(s)


def config_store_query_to_dict(x: ConfigStoreQuery) -> Any:
    return to_class(ConfigStoreQuery, x)


def config_store_contents_element_from_dict(s: Any) -> ConfigStoreContentsElement:
    return ConfigStoreContentsElement.from_dict(s)


def config_store_contents_element_to_dict(x: ConfigStoreContentsElement) -> Any:
    return to_class(ConfigStoreContentsElement, x)


def config_store_contents_from_dict(s: Any) -> List[ConfigStoreContentsElement]:
    return from_list(ConfigStoreContentsElement.from_dict, s)


def config_store_contents_to_dict(x: List[ConfigStoreContentsElement]) -> Any:
    return from_list(lambda x: to_class(ConfigStoreContentsElement, x), x)



from .generated import (
    Cfgu,
    CfguType,
    Config,
    ConfigSchemaType,
    ConfigStoreQuery,
)
from .command import Command
from .config_schema import ConfigSchema
from .config_set import ConfigSet
from .config_store import ConfigStore

__all__ = [
    "Cfgu",
    "CfguType",
    "Config",
    "ConfigSchemaType",
    "ConfigStoreQuery",
    "Command",
    "ConfigSchema",
    "ConfigSet",
    "ConfigStore",
]



from google.cloud.secretmanager import SecretManagerServiceClient
from google.cloud.secretmanager_v1 import (
    Secret,
    SecretPayload,
)

from .key_value_store import KeyValueConfigStore


class GCPSecretManagerConfigStore(KeyValueConfigStore):
    """A `ConfigStore` persisted in GCP Secret Manager"""

    _client: SecretManagerServiceClient
    _project_id: str

    def __init__(self, project_id: str, **client_kwargs) -> None:
        super().__init__(type="gcp-secret-manager")
        self._project_id = project_id
        self._client = SecretManagerServiceClient(**client_kwargs)

    def _format_key(self, key: str) -> str:
        return f"projects/{self._project_id}/secrets/{key}"

    def get_by_key(self, key: str) -> str:
        response = self._client.access_secret_version(
            name=f"{self._format_key(key)}/versions/latest"
        )
        return response.payload.data.decode("utf-8")

    def _add_secret_version(self, secret_id: str, secret_data: str):
        payload = SecretPayload()
        payload.data = secret_data.encode("utf-8")
        self._client.add_secret_version(parent=secret_id, payload=payload)

    def upsert(self, key: str, value: str):
        formatted_key = self._format_key(key)
        secret = Secret()
        secret.replication.automatic = {}
        try:
            self._client.create_secret(
                parent=f"projects/{self._project_id}",
                secret_id=key,
                secret=secret,
            )
        except (Exception,):
            pass
        self._add_secret_version(formatted_key, value)

    def delete(self, key: str):
        self._client.delete_secret(name=self._format_key(key))
